def verify_endpoint(endpoint, df):
    from os import path, getcwd
    import pandas as pd
    from json import loads
    endpoint_name = endpoint.lower().replace("/", "_")

    default_json_file = f"{endpoint_name}.json"

    __location__ = path.realpath(path.join(getcwd(), path.dirname(__file__)))
    file_path = f"../defaults/types/{default_json_file}"
    type_file = open(path.join(__location__, file_path),'r')
    default_type = loads(type_file.read())
    type_file.close()

    default_df = pd.json_normalize(default_type, sep="_")

    df_columns_list = df.columns.values.tolist()
    default_df_columns_list = default_df.columns.values.tolist()
    # print(df_columns_list)
    # print(default_df_columns_list)

    error_count = 0
    for column in df_columns_list:
        try:
            default_df_columns_list.index(column)
        except Exception as e:
            print(f"Error trying to find {column} in default for endpoint {endpoint_name}. Please update JSON to include column. ")
            error_count += 1
    print(f"Endpoint {endpoint_name} has been verified. There are a total of {error_count} issues.")

    return default_df

def titanpy_dataframe(endpoint_list, st_creds_path):
    
    from Titanpy import Titanpy
    import pandas as pd

    tp = Titanpy()
    tp.Connect(cred_path = st_creds_path)

    if isinstance(endpoint_list, list):
        pass
    else:
        endpoint_list = [endpoint_list]

    df_dict = {}
    for endpoint in endpoint_list:

        result = tp.Get(endpoint)
        result_json = result.json()

        data = result_json['data']
        has_more = result_json['hasMore']
        continue_from = result_json['continueFrom']
        print(has_more)

        df = pd.json_normalize(data, sep="_")
        verify_endpoint(endpoint, df)

        df["tenant_id"] = tp.tenant_id
        df_dict[endpoint] = df

    print(df_dict['export/jobs'])


    


# This script contains functions that Atlas uses to created different types of source datasets and initializes the database using sqlalchemy.

def Default(sql_creds_path, st_creds_path):
    
    from database_tools import create_engine, run_sql
    from titanpy_tools import titanpy_dataframe

    endpoint_list = [
        'export/jobs',
        'export/customers'
    ]

    titanpy_dataframe(endpoint_list, st_creds_path)

    # engine = create_engine(sql_creds_path)
    # run_sql(engine, '../sql/create_atlas_schema.sql')
    # run_sql(engine, "../sql/create_test_table.sql")


if __name__ == "__main__":
    Default("./credentials/postgresql_credentials.json", "./credentials/servicetitan_credentials.json")



def create_engine(creds_file_path):

    from sqlalchemy import create_engine
    from json import load

    try:
        creds_json = open(creds_file_path)
        creds_data = load(creds_json)
        creds_json.close()
    except Exception as e:
        print("There was an error with the requested credential path.")
        print(e)

    if creds_data['db_type'] == 'postgresql15':
        connection_string = f"postgresql://{creds_data['user']}:{creds_data['pass']}@{creds_data['host']}:{creds_data['port']}/{creds_data['dbname']}"
    
    engine = create_engine(connection_string)

    return engine


def run_sql(engine, file_path):
    # Will run a sql file starting from the files base directory.
    from os import path, getcwd
    from sqlalchemy import text

    __location__ = path.realpath(path.join(getcwd(), path.dirname(__file__)))
    sql_file = open(path.join(__location__, file_path),'r')
    sql = sql_file.read()
    sql_file.close()
    sqlcommands = sql.split(';')

    while("" in sqlcommands):
        sqlcommands.remove("")

    with engine.connect() as conn:
        if len(sqlcommands)>1:
            for command in sqlcommands:
                conn.execute(text(command))
            conn.commit()
        else:
            result = conn.execute(text(sql))
            conn.commit()
            return result





def get_request(credentials, query, url):
    
    import requests

    auth = {
        "ST-App-Key": credentials["APP_KEY"],
        "Authorization": credentials["ACCESS_TOKEN"]
    }
    print(f"requesting from {url}")
    default_query_parameters = {
    }
    if query != None:
        for request in query:
            default_query_parameters[request] = query[request]
    try:
        request = requests.get(url=url,params=default_query_parameters, headers=auth)
        return request
    except Exception as e:
        print(f"There was an error getting data from {url}")
        print(e)

def get(credentials, endpoint, query, id, category, *args, **kwargs):

    general_urls = {
        "url": "https://api.servicetitan.io/",
        "accounting_url_tenant": f"https://api.servicetitan.io/accounting/v2/tenant/{credentials['TENANT_ID']}/",
        "reporting_url_tenant": f"https://api.servicetitan.io/reporting/v2/tenant/{credentials['TENANT_ID']}/",
        "forms_url_tenant":f"https://api.servicetitan.io/forms/v2/tenant/{credentials['TENANT_ID']}/",
        "crm_url_tenant":f"https://api.servicetitan.io/crm/v2/tenant/{credentials['TENANT_ID']}/",
        "dispatch_url_tenant":f"https://api.servicetitan.io/dispatch/v2/tenant/{credentials['TENANT_ID']}/",
        "installed_systems_url_tenant":f"https://api.servicetitan.io/equipmentsystems/v2/tenant/{credentials['TENANT_ID']}/",
        "inventory_url_tenant":f"https://api.servicetitan.io/inventory/v2/tenant/{credentials['TENANT_ID']}/",
        "job_booking_url_tenant":f"https://api.servicetitan.io/jbce/v2/tenant/{credentials['TENANT_ID']}/",
        "job_planning_url_tenant": f"https://api.servicetitan.io/jpm/v2/tenant/{credentials['TENANT_ID']}/",
        "marketing_url_tenant": f"https://api.servicetitan.io/marketing/v2/tenant/{credentials['TENANT_ID']}/",
        "marketing_reputation_url_tenant": f"https://api.servicetitan.io/marketingreputation/v2/tenant/{credentials['TENANT_ID']}/",
        "memberships_url_tenant": f"https://api.servicetitan.io/memberships/v2/tenant/{credentials['TENANT_ID']}/",
        "payroll_url_tenant": f"https://api.servicetitan.io/payroll/v2/tenant/{credentials['TENANT_ID']}/",
        "pricebook_url_tenant": f"https://api.servicetitan.io/pricebook/v2/tenant/{credentials['TENANT_ID']}/",
        "estimates_url_tenant": f"https://api.servicetitan.io/sales/v2/tenant/{credentials['TENANT_ID']}/",
        "service_agreements_url_tenant": f"https://api.servicetitan.io/service-agreements/v2/tenant/{credentials['TENANT_ID']}/",
        "settings_url_tenant": f"https://api.servicetitan.io/settings/v2/tenant/{credentials['TENANT_ID']}/",
        "task_management_url_tenant": f"https://api.servicetitan.io/taskmanagement/v2/tenant/{credentials['TENANT_ID']}/",
        "telecom_url_tenant": f"https://api.servicetitan.io/telecom/v2/tenant/{credentials['TENANT_ID']}/",
        "telecom_url_v3_tenant": f"https://api.servicetitan.io/telecom/v3/tenant/{credentials['TENANT_ID']}/",
    }

    # --------- ENDPOINT GROUPS --------- #

    # Accounting
    accounting_endpoints = [
        'export/inventory-bills',
        'export/invoice-items',
        'export/invoices',
        'export/payments',
        'ap-credits',
        'ap-payments',
        'inventory-bills',
        'invoices',
        'journal-entries',
        'payments',
        'payment-terms',
        'tax-zones',
    ]
    accounting_id_endpoints = [
        'journal-entries-details/',
        'journal-entries-summary/',
        'payment-types/',
    ]

    # CRM
    crm_endpoints = [
        'export/bookings',
        'export/customers',
        'export/customers/contacts',
        'export/leads',
        'export/locations',
        'export/locations/contacts',
        'booking-provider-tags',
        'bookings',
        'customers',
        'customers/contacts',
        'leads',
        'locations',
        'locations/contacts',
    ]
    crm_id_endpoints = [
        'booking-provider-tags/',
        '/bookings',
        '/bookings/',
        '/bookings-contacts/',
        'bookings/',
        'bookings-contacts/',
        'customers/',
        'customers-contacts/',
        'customers-notes/',
        'leads/',
        'leads-notes/',
        'locations/',
        'locations-contacts/',
        'locations-notes/',
    ]

    # Dispatch
    dispatch_endpoints = [
        'export/appointment-assignments',
        'appointment_assignments',
        'non-job-appointments',
        'teams',
        'technician-shifts',
        'zones',
    ]
    dispatch_id_endpoints = [
        'non-job-appointments/',
        'teams/',
        'technician-shifts/',
        'zones/',
    ]

    # Installed Systems
    installed_systems_endpoints = [
        'installed-equipment',
    ]
    installed_systems_id_endpoints = [
        'installed-equipment/',
    ]

    # Forms
    forms_endpoints = [
        'forms',
        'submissions',
    ]

    # Inventory
    inventory_endpoints = [
        'export/purchase-orders',
        'adjustments',
        'purchase-orders',
        'purchase-order-markups',
        'purchase-order-types',
        'receipts',
        'returns',
        'transfers',
        'trucks',
        'vendors',
        'warehouses',
    ]
    inventory_id_endpoints = [
        'purchase-orders/',
        'purchase-order-markups/',
        'vendors/',
    ]

    # Job Booking
    job_booking_endpoints = [
        'call-reasons',
    ]

    # Job Planning and Management
    job_planning_endpoints = [
        'export/appointments',
        'export/job-canceled-logs',
        'export/jobs',
        'export/projects',
        'appointments',
        'job-cancel-reasons',
        'job-hold-reasons',
        'jobs',
        'job-types',
        'projects',
        'project-statuses',
        'project-substatuses',
    ]
    job_planning_id_endpoints = [
        'appointments/',
        'jobs-cancel-reasons//',
        'jobs/',
        'jobs-history/',
        'jobs-notes/',
        'job-types/',
        'projects/',
        'projects-notes/',
        'project-statuses/',
        'project-substatuses/',
    ]

    # Marketing
    marketing_endpoints = [
        'categories',
        'costs',
        'campaigns',
        'suppressions',
    ]
    marketing_id_endpoints = [
        'categories/',
        'costs/',
        'campaigns/',
        'campaigns-costs/',
        'suppressions/',
    ]

    # Marketing Reputation
    marketing_reputation_endpoints = [
        'reviews',
    ]

    # Memberships Endpoints
    memberships_endpoints = [
        'export/invoice-templates',
        'export/membership-types',
        'export/memberships',
        'export/recurring-service-events',
        'export/recurring-service-types',
        'export/recurring-services',
        'memberships',
        'recurring-service-events',
        'recurring-services',
        'membership-types',
        'recurring-service-types',

    ]
    memberships_id_endpoints = [
        'memberships/',
        'memberships-status-changes/',
        'invoice-templates/',
        'invoice-templates//',
        'recurring-services/',
        'membership-types/',
        'membership-types-discounts/',
        'membership-types-duration-billing-items/',
        'membership-types-recurring-service-items/',
        'recurring-service-types/',
    ]

    # Payroll
    payroll_endpoints = [
        'export/activity-codes',
        'export/gross-pay-items',
        'export/jobs/splits',
        'export/jobs/timesheets',
        'export/payroll-adjustments',
        'export/timesheet-codes',
        'activity-codes',
        'gross-pay-items',
        'jobs/splits',
        'locations/rates',
        'payroll-adjustments',
        'payrolls',
        'timesheet-codes',
        'jobs/timesheets',
        'non-job-timesheets',
    ]
    payroll_id_endpoints = [
        'activity-codes/',
        'jobs-splits/',
        'payroll-adjustments/',
        'employees-payrolls/',
        'technicians-payrolls/',
        'timesheet-codes/',
        'jobs-timesheets/',
    ]

    # Pricebook
    pricebook_endpoints = [
        'categories',
        'discounts-and-fees',
        'equipment',
        'images',
        'materials',
        'materialsmarkup',
        'services',
    ]
    pricebook_id_endpoints = [
        'categories/',
        'discounts-and-fees/',
        'equipment/',
        'materials/',
        'materialsmarkup/',
        'services/',
    ]

    # Reporting
    reporting_endpoints = [
        'report-categories',
    ]
    reporting_id_endpoints = [
        'dynamic-value-sets/',
        '/reports',
        '/reports/',
    ]

    # Sales/Estimates
    estimates_endpoints = [
        'estimates/export',
        'estimates',
        'estimates/items',
    ]
    estimates_id_endpoints = [
        'estimates/',
    ]

    # Service Agreements
    service_agreements_endpoints = [
        'export/service-agreements',
        'service-agreements',
    ]
    service_agreements_id_endpoints = [
        'service-agreements/',
    ]

    # Settings
    settings_endpoints = [
        'export/business-units',
        'export/employees',
        'export/tag-types',
        'export/technicians',
        'business_units',
        'employees',
        'tag-types',
        'technicians',
        'user-roles',
    ]
    settings_id_endpoints = [
        'business_units/',
        'employees/',
        'technicians/',
    ]

    # Task Management
    task_management_endpoints = [
        'data',
    ]

    # Telecom
    telecom_endpoints = [
        'export/calls',
    ]
    telecom_v3_endpoints = [
        'calls',
    ]
    telecom_id_endpoints = [
        'calls/',
        'calls-recording/',
        'calls-voicemail/',
    ]

    # --------- AVAILABLE ENDPOINTS --------- #

    available_endpoints = [
    ]
    available_endpoint_groups = [
        accounting_id_endpoints,
        reporting_endpoints,
        reporting_id_endpoints,
        forms_endpoints,
        crm_endpoints,
        crm_id_endpoints,
        dispatch_endpoints,
        dispatch_id_endpoints,
        installed_systems_endpoints,
        installed_systems_id_endpoints,
        inventory_endpoints,
        inventory_id_endpoints,
        job_booking_endpoints,
        job_planning_endpoints,
        job_planning_id_endpoints,
        marketing_endpoints,
        marketing_id_endpoints,
        marketing_reputation_endpoints,
        memberships_endpoints,
        memberships_id_endpoints,
        payroll_endpoints,
        payroll_id_endpoints,
        pricebook_endpoints,
        pricebook_id_endpoints,
        estimates_endpoints,
        estimates_id_endpoints,
        service_agreements_endpoints,
        service_agreements_id_endpoints,
        settings_endpoints,
        settings_id_endpoints,
        task_management_endpoints,

    ]
    for group in available_endpoint_groups:
        available_endpoints.extend(group)

    # --------- ENDPOINT TESTS --------- #

    # Tests if endpoint has been manually added or not.
    if endpoint in available_endpoints:

        # Accounting Endpoints

        if endpoint in accounting_endpoints:
            url = f"{general_urls['accounting_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        if endpoint in accounting_id_endpoints:
            if endpoint == 'journal-entries-details/':
                if id != None:
                    url = f"{general_urls['accounting_url_tenant']}journal-entries/{id}/details"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint == 'journal-entries-summary/':
                if id != None:
                    url = f"{general_urls['accounting_url_tenant']}journal-entries/{id}/summary"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint == 'payment-types/':
                if id != None:
                    url = f"{general_urls['accounting_url_tenant']}payment-types/{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
        
        # CRM Endpoints
        
        if endpoint in crm_endpoints:
            url = f"{general_urls['crm_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        if endpoint in crm_id_endpoints:
            
            if endpoint in ['booking-provider-tags/', 'bookings/', 'customers/','leads/','locations/']:
                if id != None:
                    url = f"{general_urls['crm_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

            if endpoint in ['/bookings']:
                if category != None:
                    url = f"{general_urls['crm_url_tenant']}{category}{endpoint}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an Category in order to run. Please enter category as an arg.")
            
            if endpoint in ['/bookings/',]:
                if category != None and id != None:
                    url = f"{general_urls['crm_url_tenant']}{category}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID and a Category in order to run. Please enter id and category as an arg.")

            if endpoint == '/bookings-contacts/':
                if category != None and id != None:
                    url = f"{general_urls['crm_url_tenant']}{category}/bookings/{id}/contacts"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID and a Category in order to run. Please enter id and category as an arg.")

            if endpoint == 'bookings-contacts/':
                if id != None:
                    url = f"{general_urls['crm_url_tenant']}bookings/{id}/contacts"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

            if endpoint == 'customers-contacts/':
                if id != None:
                    url = f"{general_urls['crm_url_tenant']}customers/{id}/contacts"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

            if endpoint == 'customers-notes/':
                if id != None:
                    url = f"{general_urls['crm_url_tenant']}customers/{id}/notes"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

            if endpoint == 'leads-notes/':
                if id != None:
                    url = f"{general_urls['crm_url_tenant']}leads/{id}/notes"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

            if endpoint == 'locations-contacts/':
                if id != None:
                    url = f"{general_urls['crm_url_tenant']}locations/{id}/contacts"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

            if endpoint == 'locations-notes/':
                if id != None:
                    url = f"{general_urls['crm_url_tenant']}locations/{id}/notes"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Dispatch Endpoints

        if endpoint in dispatch_endpoints:
            url = f"{general_urls['dispatch_url_tenant']}{endpoint}"
            return get_request(credentials,query,url)
        
        if endpoint in ['non-job-appointments/','teams/','technician-shifts/','zones/']:
            if id != None:
                url = f"{general_urls['dispatch_url_tenant']}{endpoint}{id}"
                return get_request(credentials,query,url)
            else:
                print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Installed Systems Endpoints
        
        if endpoint in installed_systems_endpoints:
            url = f"{general_urls['installed_systems_url_tenant']}{endpoint}"
            return get_request(credentials,query,url)
        
        if endpoint in ['installed-equipment/',]:
            if id != None:
                url = f"{general_urls['installed_systems_url_tenant']}{endpoint}{id}"
                return get_request(credentials,query,url)
            else:
                print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Forms Endpoints
        
        if endpoint in forms_endpoints:
            url = f"{general_urls['forms_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        # Inventory Endpoints

        if endpoint in inventory_endpoints:
            url = f"{general_urls['inventory_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        if endpoint in ['purchase-orders/','purchase-order-markups/','vendors/']:
            if id != None:
                url = f"{general_urls['inventory_url_tenant']}{endpoint}{id}"
                return get_request(credentials,query,url)
            else:
                print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Job Booking Endpoints

        if endpoint in job_booking_endpoints:
            url = f"{general_urls['job_booking_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        # Job Planning Endpoints

        if endpoint in job_planning_endpoints:
            url = f"{general_urls['job_planning_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)       

        if endpoint in job_planning_id_endpoints:
            
            if endpoint in ['appointments/','jobs/','job-types/','projects/','project-statuses/','project-substatuses/']:
                if id != None:
                    url = f"{general_urls['job_planning_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['job-cancel-reasons//']:
                if id != None:
                    url = f"{general_urls['job_planning_url_tenant']}jobs/cancel-reasons?ids={id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires IDs in order to run. Please enter ids as an arg.")
            if endpoint in ['jobs-history/']:
                if id != None:
                    url = f"{general_urls['job_planning_url_tenant']}jobs/{id}/history"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['jobs-notes/']:
                if id != None:
                    url = f"{general_urls['job_planning_url_tenant']}jobs/{id}/notes"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['projects-notes/']:
                if id != None:
                    url = f"{general_urls['job_planning_url_tenant']}projects/{id}/notes"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Marketing Endpoints

        if endpoint in marketing_endpoints:
            url = f"{general_urls['marketing_url_tenant']}{endpoint}"
            return get_request(credentials, query, url) 
        
        if endpoint in marketing_id_endpoints:
            if endpoint in ['categories/','costs/','campaigns/','suppressions/']:
                if id != None:
                    url = f"{general_urls['marketing_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg. (Email is considered and ID for suppressions/ endpoints.)")
            if endpoint in ['campaigns-costs/']:
                if id != None:
                    url = f"{general_urls['marketing_url_tenant']}campaigns/{id}/costs"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
        
        # Marketing Reputation Endpoints

        if endpoint in marketing_reputation_endpoints:
            url = f"{general_urls['marketing_reputation_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        # Memberships Endpoints

        if endpoint in memberships_endpoints:
            url = f"{general_urls['memberships_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        
        if endpoint in memberships_id_endpoints:
            if endpoint in ['memberships/', 'invoice-templates/', 'recurring-services/', 'membership-types/', 'recurring-service-types/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['memberships-status-changes/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}memberships/{id}/status-changes"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['invoice-templates//']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}invoice-templates?id={id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires IDs in order to run. Please enter ids as an arg.")
            if endpoint in ['membership-types-discounts/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}membership-types/{id}/discounts"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['membership-types-duration-billing-items/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}membership-types/{id}/duration-billing-items"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['membership-types-recurring-service-items/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}membership-types/{id}/recurring-service-items"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Payroll Endpoints

        if endpoint in payroll_endpoints:
            url = f"{general_urls['payroll_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        
        if endpoint in payroll_id_endpoints:
            if endpoint in ['activity-codes/', 'payroll-adjustments/', 'timesheet-codes/',]:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['jobs-splits/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}jobs/{id}/splits"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['employees-payrolls/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}employees/{id}/payrolls"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['technicians-payrolls/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}technicians/{id}/payrolls"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['jobs-timesheets/']:
                if id != None:
                    url = f"{general_urls['memberships_url_tenant']}jobs/{id}/timesheets"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Pricebook Endpoints

        if endpoint in pricebook_endpoints:
            url = f"{general_urls['pricebook_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        if endpoint in pricebook_id_endpoints:
            if endpoint in ['categories/', 'discounts-and-fees/', 'equipment/', 'materials/', 'materialsmarkup/', 'services/',]:
                if id != None:
                    url = f"{general_urls['pricebook_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Reporting Endpoints

        if endpoint in reporting_endpoints:
            url = f"{general_urls['reporting_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        
        if endpoint in reporting_id_endpoints:
            if endpoint == 'dynamic-value-sets/':
                if id != None:
                    url = f"{general_urls['reporting_url_tenant']}dynamic-value-sets/{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint == '/reports':
                if category != None:
                    url = f"{general_urls['reporting_url_tenant']}report-category/{category}/reports"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an Category in order to run. Please enter category as an arg.")
            if endpoint == '/reports/':
                if id != None and category != None:
                    url = f"{general_urls['reporting_url_tenant']}report-category/{category}/reports/{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID and a Category in order to run. Please enter id and category as an arg.")

        # Sales/Estimates Endpoints

        if endpoint in estimates_endpoints:
            url = f"{general_urls['estimates_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        
        if endpoint in estimates_id_endpoints:
            if endpoint in ['estimates/',]:
                if id != None:
                    url = f"{general_urls['estimates_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
        
        # Service Agreement Endpoints

        if endpoint in service_agreements_endpoints:
            url = f"{general_urls['service_agreements_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        
        if endpoint in service_agreements_id_endpoints:
            if endpoint in ['service-agreements/',]:
                if id != None:
                    url = f"{general_urls['service_agreements_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Settings Endpoints

        if endpoint in settings_endpoints:
            url = f"{general_urls['settings_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        
        if endpoint in settings_id_endpoints:
            if endpoint in ['business_units/', 'employees/', 'technicians/',]:
                if id != None:
                    url = f"{general_urls['settings_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

        # Task Management Endpoints

        if endpoint in task_management_endpoints:
            url = f"{general_urls['task_management_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)

        # Telecom Endpoints

        if endpoint in telecom_endpoints:
            url = f"{general_urls['telecom_url_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        if endpoint in telecom_v3_endpoints:
            url = f"{general_urls['telecom_url_v3_tenant']}{endpoint}"
            return get_request(credentials, query, url)
        if endpoint in telecom_id_endpoints:
            if endpoint in ['calls/',]:
                if id != None:
                    url = f"{general_urls['telecom_url_tenant']}{endpoint}{id}"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['calls-recording/',]:
                if id != None:
                    url = f"{general_urls['telecom_url_tenant']}calls/{id}/recording"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")
            if endpoint in ['calls-voicemail/',]:
                if id != None:
                    url = f"{general_urls['telecom_url_tenant']}calls/{id}/voicemail"
                    return get_request(credentials,query,url)
                else:
                    print("The requested endpoint requires an ID in order to run. Please enter id as an arg.")

    # Attempts to make a request to a general endpoint if it has not been manually added
    # TO-DO AI assisted url creation
    else:
        try:
            url = f"https://api.servicetitan.io/jpm/v2/tenant/{credentials['TENANT_ID']}/{endpoint}"
            return get_request(credentials, query, url)
        except Exception as e:
            print(f"There was an error getting data from {endpoint}. The endpoint was not found in our current available\
                  endpoint list and could not be interpretted automatically. Please submit a request for this endpoint.")
            print(e)



# class Atlas:

#     def __init__(self, tp=None):
#         self.tp = tp

#     def Build(self, type, db_cred_path, st_credential_paths*):

#         from Titanpy.titanpy import Titanpy

#         # Connect to database
#         import sqlalchemy
#         # Create Schema
#         # Loop through connections and build database
#         self.tp = Titanpy()
#         self.tp.Connect(st_credential_paths)

#     def Update(self, type, db_cred_path, st_credential_paths*):
#         pass





from Titanpy import Titanpy
# --------- ENDPOINT GROUPS --------- #

# Accounting
accounting_endpoints = [
    'export/inventory-bills',
    'export/invoice-items',
    'export/invoices',
    'export/payments',
    'ap-credits',
    'ap-payments',
    'inventory-bills',
    'invoices',
    'journal-entries',
    'payments',
    'payment-terms',
    'tax-zones',
]
accounting_id_endpoints = [
    'journal-entries-details/',
    'journal-entries-summary/',
    'payment-types/',
]

# CRM
crm_endpoints = [
    'export/bookings',
    'export/customers',
    'export/customers/contacts',
    'export/leads',
    'export/locations',
    'export/locations/contacts',
    'booking-provider-tags',
    'bookings',
    'customers',
    'customers/contacts',
    'leads',
    'locations',
    'locations/contacts',
]
crm_id_endpoints = [
    'booking-provider-tags/',
    '/bookings',
    '/bookings/',
    '/bookings-contacts/',
    'bookings/',
    'bookings-contacts/',
    'customers/',
    'customers-contacts/',
    'customers-notes/',
    'leads/',
    'leads-notes/',
    'locations/',
    'locations-contacts/',
    'locations-notes/',
]

# Dispatch
dispatch_endpoints = [
    'export/appointment-assignments',
    'appointment_assignments',
    'non-job-appointments',
    'teams',
    'technician-shifts',
    'zones',
]
dispatch_id_endpoints = [
    'non-job-appointments/',
    'teams/',
    'technician-shifts/',
    'zones/',
]

# Installed Systems
installed_systems_endpoints = [
    'installed-equipment',
]
installed_systems_id_endpoints = [
    'installed-equipment/',
]

# Forms
forms_endpoints = [
    'forms',
    'submissions',
]

# Inventory
inventory_endpoints = [
    'export/purchase-orders',
    'adjustments',
    'purchase-orders',
    'purchase-order-markups',
    'purchase-order-types',
    'receipts',
    'returns',
    'transfers',
    'trucks',
    'vendors',
    'warehouses',
]
inventory_id_endpoints = [
    'purchase-orders/',
    'purchase-order-markups/',
    'vendors/',
]

# Job Booking
job_booking_endpoints = [
    'call-reasons',
]

# Job Planning and Management
job_planning_endpoints = [
    'export/appointments',
    'export/job-canceled-logs',
    'export/jobs',
    'export/projects',
    'appointments',
    'job-cancel-reasons',
    'job-hold-reasons',
    'jobs',
    'job-types',
    'projects',
    'project-statuses',
    'project-substatuses',
]
job_planning_id_endpoints = [
    'appointments/',
    'jobs-cancel-reasons//',
    'jobs/',
    'jobs-history/',
    'jobs-notes/',
    'job-types/',
    'projects/',
    'projects-notes/',
    'project-statuses/',
    'project-substatuses/',
]

# Marketing
marketing_endpoints = [
    'categories',
    'costs',
    'campaigns',
    'suppressions',
]
marketing_id_endpoints = [
    'categories/',
    'costs/',
    'campaigns/',
    'campaigns-costs/',
    'suppressions/',
]

# Marketing Reputation
marketing_reputation_endpoints = [
    'reviews',
]

# Memberships Endpoints
memberships_endpoints = [
    'export/invoice-templates',
    'export/membership-types',
    'export/memberships',
    'export/recurring-service-events',
    'export/recurring-service-types',
    'export/recurring-services',
    'memberships',
    'recurring-service-events',
    'recurring-services',
    'membership-types',
    'recurring-service-types',

]
memberships_id_endpoints = [
    'memberships/',
    'memberships-status-changes/',
    'invoice-templates/',
    'invoice-templates//',
    'recurring-services/',
    'membership-types/',
    'membership-types-discounts/',
    'membership-types-duration-billing-items/',
    'membership-types-recurring-service-items/',
    'recurring-service-types/',
]

# Payroll
payroll_endpoints = [
    'export/activity-codes',
    'export/gross-pay-items',
    'export/jobs/splits',
    'export/jobs/timesheets',
    'export/payroll-adjustments',
    'export/timesheet-codes',
    'activity-codes',
    'gross-pay-items',
    'jobs/splits',
    'locations/rates',
    'payroll-adjustments',
    'payrolls',
    'timesheet-codes',
    'jobs/timesheets',
    'non-job-timesheets',
]
payroll_id_endpoints = [
    'activity-codes/',
    'jobs-splits/',
    'payroll-adjustments/',
    'employees-payrolls/',
    'technicians-payrolls/',
    'timesheet-codes/',
    'jobs-timesheets/',
]

# Pricebook
pricebook_endpoints = [
    'categories',
    'discounts-and-fees',
    'equipment',
    'images',
    'materials',
    'materialsmarkup',
    'services',
]
pricebook_id_endpoints = [
    'categories/',
    'discounts-and-fees/',
    'equipment/',
    'materials/',
    'materialsmarkup/',
    'services/',
]

# Reporting
reporting_endpoints = [
    'report-categories',
]
reporting_id_endpoints = [
    'dynamic-value-sets/',
    '/reports',
    '/reports/',
]

# Sales/Estimates
estimates_endpoints = [
    'estimates/export',
    'estimates',
    'estimates/items',
]
estimates_id_endpoints = [
    'estimates/',
]

# Service Agreements
service_agreements_endpoints = [
    'export/service-agreements',
    'service-agreements',
]
service_agreements_id_endpoints = [
    'service-agreements/',
]

# Settings
settings_endpoints = [
    'export/business-units',
    'export/employees',
    'export/tag-types',
    'export/technicians',
    'business_units',
    'employees',
    'tag-types',
    'technicians',
    'user-roles',
]
settings_id_endpoints = [
    'business_units/',
    'employees/',
    'technicians/',
]

# Task Management
task_management_endpoints = [
    'data',
]

# Telecom
telecom_endpoints = [
    'export/calls',
]
telecom_v3_endpoints = [
    'calls',
]
telecom_id_endpoints = [
    'calls/',
    'calls-recording/',
    'calls-voicemail/',
]

# --------- AVAILABLE ENDPOINTS --------- #

available_endpoints = [
]
available_endpoint_groups = [
    accounting_id_endpoints,
    reporting_endpoints,
    reporting_id_endpoints,
    forms_endpoints,
    crm_endpoints,
    crm_id_endpoints,
    dispatch_endpoints,
    dispatch_id_endpoints,
    installed_systems_endpoints,
    installed_systems_id_endpoints,
    inventory_endpoints,
    inventory_id_endpoints,
    job_booking_endpoints,
    job_planning_endpoints,
    job_planning_id_endpoints,
    marketing_endpoints,
    marketing_id_endpoints,
    marketing_reputation_endpoints,
    memberships_endpoints,
    memberships_id_endpoints,
    payroll_endpoints,
    payroll_id_endpoints,
    pricebook_endpoints,
    pricebook_id_endpoints,
    estimates_endpoints,
    estimates_id_endpoints,
    service_agreements_endpoints,
    service_agreements_id_endpoints,
    settings_endpoints,
    settings_id_endpoints,
    task_management_endpoints,

]
for group in available_endpoint_groups:
    available_endpoints.extend(group)



# --------- TEST ENDPOINTS --------- #
tp = Titanpy()
tp.Connect(cred_path="./credentials/servicetitan_credentials.json")
for endpoint in available_endpoints:
    response = tp.Get(endpoint = endpoint, id=1, category='gamer')
    try:
        print(response.status_code)
        print(response.reason)
    except:
        response = tp.Get(endpoint = endpoint, id=1, category='gamer')
        try:
            print(response.status_code)
            print(response.reason)
        except:
            pass


# Titan Class Doc
class Titanpy:

    # Variable Initialization
    def __init__(self, credentials=None, client_id=None, client_secret=None, app_id=None, app_key=None, tenant_id=None, access_token=None):
        
        # credential variables
        self.credentials = credentials
        self.client_id = client_id
        self.client_secret = client_secret
        self.app_id = app_id
        self.app_key = app_key
        self.tenant_id = tenant_id
        self.access_token = access_token

    # Connect method takes credentials and creates a connection token in order to request from other API sources
    def Connect(self, cred_path):

        from Titanpy.connect import load_credentials

        credentials = load_credentials(cred_path)

        self.client_id = credentials["CLIENT_ID"]
        self.client_secret = credentials["CLIENT_SECRET"]
        self.app_id = credentials["APP_ID"]
        self.app_key = credentials["APP_KEY"]
        self.tenant_id = credentials["TENANT_ID"]
        self.timezone = credentials["TIMEZONE"]
        self.access_token = credentials["ACCESS_TOKEN"]
        self.credentials = credentials

    # Returns data from a source.
    # Use Connect() before using API methods.
    def Get(self, endpoint=None, query = None, id=None, category=None, url=None, *args, **kwargs):

        if self.access_token == None:
            print("No proper connection. Please ensure Connect method has been run successfully.")
        elif endpoint !=None:
            from Titanpy.get import get
            print(f"Endpoint received. Requesting for endpoint ({endpoint})")
            return get(credentials = self.credentials, endpoint = endpoint, query = query, id=id, category=category)
        elif url !=None:
            from Titanpy.get import get_request
            print(f"Url received. Requesting from url ({url}).")
            return get_request(credentials = self.credentials, query=query, url=url)
        else:
            print("No endpoint or url has been entered. Please enter an endpoint or a url.")

    def Post(self):

        if self.access_token == None:
            print("No proper connection. Please ensure Connect method has been run successfully.")

        else:
            print("This method has not been coded yet. See https://github.com/pabboat/Titanpy for more information.")

    def Del(self):

        if self.access_token == None:
            print("No proper connection. Please ensure Connect method has been run successfully.")

        else:
            print("This method has not been coded yet. See https://github.com/pabboat/Titanpy for more information.")

    def Put(self):

        if self.access_token == None:
            print("No proper connection. Please ensure Connect method has been run successfully.")

        else:
            print("This method has not been coded yet. See https://github.com/pabboat/Titanpy for more information.")

    def Patch(self):

        if self.access_token == None:
            print("No proper connection. Please ensure Connect method has been run successfully.")

        else:
            print("This method has not been coded yet. See https://github.com/pabboat/Titanpy for more information.")



# visit my page! happy to help: https://github.com/pabboat 
# I recommend using "from titanpy import Titanpy as tp"

# imports
from Titanpy.titanpy import Titanpy
# from Titanpy.atlas import Atlas

# test scripts

if __name__ == "__main__":
    b = Titanpy()

    b.Connect(cred_path="./credentials/servicetitan_credentials.json")

    print(b.Get(endpoint = 'receipts'))
